#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import collections
import io
import struct
import sys

FMT_VERSION = "BB"

GEOB_KEY = "Serato BeatGrid"

NonTerminalBeatgridMarker = collections.namedtuple(
    "NonTerminalBeatgridMarker",
    ("position", "beats_till_next_marker"),
)
TerminalBeatgridMarker = collections.namedtuple(
    "TerminalBeatgridMarker",
    ("position", "bpm"),
)

Footer = collections.namedtuple("Footer", ("unknown",))

DataType = list[NonTerminalBeatgridMarker | TerminalBeatgridMarker | Footer]


def _check_data(data: DataType):
    nonterminal_markers: list[NonTerminalBeatgridMarker] = []
    terminal_markers: list[TerminalBeatgridMarker] = []
    footers: list[Footer] = []

    for d in data:
        if isinstance(d, NonTerminalBeatgridMarker):
            nonterminal_markers.append(d)
        elif isinstance(d, TerminalBeatgridMarker):
            terminal_markers.append(d)
        elif isinstance(d, Footer):
            footers.append(d)
        else:
            raise TypeError(f"unexpected value type {d}")

    assert (
        len(terminal_markers) == 1
    ), f"should only be 1 terminal marker, but #: {len(terminal_markers)}"
    assert len(footers) == 1, f"should only be 1 footer, but #: {len(footers)}"
    assert isinstance(data[-1], Footer), "last item should be a footer"
    assert isinstance(
        data[-2], TerminalBeatgridMarker
    ), "last item should be a terminal marker"
    return nonterminal_markers, terminal_markers, footers[0]


def parse(fp: io.BytesIO | io.BufferedReader):
    version = struct.unpack(FMT_VERSION, fp.read(2))
    assert version == (0x01, 0x00)

    num_markers = struct.unpack(">I", fp.read(4))[0]
    for i in range(num_markers):
        position = struct.unpack(">f", fp.read(4))[0]
        data = fp.read(4)
        if i == num_markers - 1:
            bpm = struct.unpack(">f", data)[0]
            yield TerminalBeatgridMarker(position, bpm)
        else:
            beats_till_next_marker = struct.unpack(">I", data)[0]
            yield NonTerminalBeatgridMarker(position, beats_till_next_marker)

    # TODO: What's the meaning of the footer byte?
    yield Footer(struct.unpack("B", fp.read(1))[0])
    assert fp.read() == b""


def write(
    data: DataType,
    fp: io.BytesIO | io.BufferedWriter,
):
    nonterminal_markers, terminal_markers, footer = _check_data(data)
    markers = nonterminal_markers + terminal_markers
    # Write version
    fp.write(struct.pack(FMT_VERSION, 0x01, 0x00))

    # Write markers
    fp.write(struct.pack(">I", len(markers)))
    for marker in markers:
        fp.write(struct.pack(">f", marker.position))
        if isinstance(marker, TerminalBeatgridMarker):
            fp.write(struct.pack(">f", marker.bpm))
        elif isinstance(marker, NonTerminalBeatgridMarker):
            fp.write(struct.pack(">I", marker.beats_till_next_marker))
        else:
            raise TypeError(f"Unexpected marker type: {type(marker)}")

    # Write footer
    fp.write(struct.pack("B", footer.unknown))


def analyze_and_write(file: str):
    import mutagen._file

    from .utils.beatgrid_analyze import analyze_beatgrid
    from .utils.tags import tag_geob

    tagfile = mutagen._file.File(file)
    assert tagfile, "file parse failed"
    bpm = float(str(tagfile["TBPM"]))

    print("Analyzing beat grid...")
    beat_analyzer = analyze_beatgrid(file, bpm_helper=bpm)

    print("Writing tags...")
    markers: DataType = [
        NonTerminalBeatgridMarker(position, 4)
        for position in beat_analyzer.downbeats[:-1]
    ] + [
        TerminalBeatgridMarker(
            beat_analyzer.downbeats[-1], bpm=bpm or beat_analyzer.bpm
        ),
        Footer(0),
    ]

    fpw = io.BytesIO()
    write(markers, fpw)
    fpw.seek(0)
    new_data = fpw.read()

    tag_geob(tagfile, GEOB_KEY, new_data)
    tagfile.save()


def main():
    import argparse

    import mutagen._file

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    args = parser.parse_args()

    tagfile = mutagen._file.File(args.file)
    if tagfile is not None:
        analyze_and_write(args.file)
    else:
        fp = open(args.file, mode="rb")

        with fp:
            for marker in parse(fp):
                print(marker)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import os
import struct
import sys
from typing import (Any, Callable, Generator, Iterable, NotRequired, Tuple,
                    TypedDict)


class DbEntry(TypedDict):
    field: str
    field_name: str
    value: str | int | bool | list["DbEntry"]
    size_bytes: int


ParsedType = Tuple[str, int, Any, bytes]


class DatabaseV2(object):
    FIELDNAMES = {
        # Database
        "vrsn": "Version",
        "otrk": "Track",
        "ttyp": "File Type",
        "pfil": "File Path",
        "tsng": "Song Title",
        "tart": "Artist",
        "talb": "Album",
        "tgen": "Genre",
        "tlen": "Length",
        "tbit": "Bitrate",
        "tsmp": "Sample Rate",
        "tsiz": "Size",
        "tbpm": "BPM",
        "tkey": "Key",
        "tart": "Artist",
        "utme": "File Time",
        "tgrp": "Grouping",
        "tlbl": "Publisher",
        "tcmp": "Composer",
        "ttyr": "Year",
        # Serato stuff
        "tadd": "Date added",
        "uadd": "Date added",
        "bbgl": "Beatgrid Locked",
        "bcrt": "Corrupt",
        "bmis": "Missing",
        # Crates
        "osrt": "Sorting",
        "brev": "Reverse Order",
        "ovct": "Column Title",
        "tvcn": "Column Name",
        "tvcw": "Column Width",
        "ptrk": "Track Path",
    }
    DEFAULT_DATABASE_FILE = os.path.join(os.path.expanduser("~"), "Music\\_Serato_\\database V2")  # type: ignore

    def __init__(self, filepath: str = DEFAULT_DATABASE_FILE):
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"file does not exist: {filepath}")

        self.filepath: str = os.path.abspath(filepath)

        self.data: Iterable[ParsedType] = self._parse(self.filepath)

    def __str__(self):
        return str(list(self.to_objects()))

    @staticmethod
    def _get_field_name(name: str):
        return DatabaseV2.FIELDNAMES.get(name, "Unknown Field")

    @staticmethod
    def _get_type(name: str) -> str:
        # vrsn field has no type_id, but contains text
        return "t" if name == "vrsn" else name[0]

    def _parse(
        self, fp: io.BytesIO | io.BufferedReader | str | None
    ) -> Generator[ParsedType]:
        if fp is None:
            fp = self.filepath
        if isinstance(fp, str):
            fp = open(fp, "rb")

        for header in iter(lambda: fp.read(8), b""):
            assert len(header) == 8
            name_ascii: bytes
            length: int
            name_ascii, length = struct.unpack(">4sI", header)
            name: str = name_ascii.decode("ascii")
            type_id: str = DatabaseV2._get_type(name)

            data = fp.read(length)
            assert len(data) == length

            value: bytes | str | tuple
            if type_id == "b":
                value = struct.unpack("?", data)[0]
            elif type_id in ("o", "r"):
                value = tuple(self._parse(io.BytesIO(data)))
            elif type_id in ("p", "t"):
                value = (data[1:] + b"\00").decode("utf-16")
            elif type_id == "s":
                value = struct.unpack(">H", data)[0]
            elif type_id == "u":
                value = struct.unpack(">I", data)[0]
            else:
                value = data

            yield name, length, value, data

    class ModifyRule(TypedDict):
        field: str
        func: Callable[[str, Any], Any]
        files: NotRequired[list[str]]

    @staticmethod
    def _modify_data(
        fp: io.BytesIO | io.BufferedWriter,
        item: Iterable[ParsedType],
        rules: list[ModifyRule] = [],
        print_changes: bool = False,
    ):
        for rule in rules:
            if "files" in rule:
                rule["files"] = [
                    os.path.normpath(os.path.splitdrive(file)[1]).lstrip("\\").upper()
                    for file in rule["files"]
                ]

        track_filename: str = ""
        for name, length, value, data in item:
            name_bytes = name.encode("ascii")
            assert len(name_bytes) == 4

            type_id: str = DatabaseV2._get_type(name)

            if name == "pfil":
                assert isinstance(value, str)
                track_filename = os.path.normpath(value)

            rule_has_been_done = False
            for rule in rules:
                if name == rule["field"] and (
                    "files" not in rule or track_filename.upper() in rule["files"]
                ):
                    maybe_new_value = rule["func"](track_filename, value)
                    if maybe_new_value is not None:
                        assert (
                            not rule_has_been_done
                        ), f"Should only pass one rule per field (field: {name})"
                        rule_has_been_done = True
                        value = maybe_new_value
                        if print_changes:
                            print(
                                f"Set {name} ({DatabaseV2._get_field_name(name)})={str(value)} in library"
                            )

            if type_id == "b":
                data = struct.pack("?", value)
            elif type_id in ("o", "r"):
                assert isinstance(value, tuple)
                nested_buffer = io.BytesIO()
                DatabaseV2._modify_data(nested_buffer, value, rules, print_changes)
                data = nested_buffer.getvalue()
            elif type_id in ("p", "t"):
                new_data = str(value).encode("utf-16")[2:]
                assert new_data[-1:] == b"\x00"
                data = data[:1] + new_data[:-1]
            elif type_id == "s":
                data = struct.pack(">H", value)
            elif type_id == "u":
                data = struct.pack(">I", value)

            length = len(data)

            header = struct.pack(">4sI", name_bytes, length)
            fp.write(header)
            fp.write(data)

    def modify_file(
        self,
        rules: list[ModifyRule],
        out_file: str | None = None,
        print_changes: bool = False,
    ):
        if out_file is None:
            out_file = self.filepath

        output = io.BytesIO()
        DatabaseV2._modify_data(output, list(self.data), rules, print_changes)

        with open(out_file, "wb") as write_file:
            write_file.write(output.getvalue())

    def to_objects(self) -> Generator[DbEntry]:
        for name, length, value, data in self.data:
            if isinstance(value, tuple):
                try:
                    new_val: list[DbEntry] = [
                        {
                            "field": n,
                            "field_name": DatabaseV2._get_field_name(n),
                            "size_bytes": l,
                            "value": v,
                        }
                        for n, l, v, d in value
                    ]
                except:
                    print(f"error on {value}")
                    raise
                value = new_val
            else:
                value = repr(value)

            yield {
                "field": name,
                "field_name": DatabaseV2._get_field_name(name),
                "size_bytes": length,
                "value": value,
            }

    def print_items(self):
        for entry in self.to_objects():
            if isinstance(entry["value"], list):
                print(
                    f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B)"
                )
                for e in entry["value"]:
                    print(
                        f"    {e['field']} ({e['field_name']}, {e['size_bytes']} B): {e['value']}"
                    )
            else:
                print(
                    f"{entry['field']} ({entry['field_name']}, {entry['size_bytes']} B): {entry['value']}"
                )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs="?", default=DatabaseV2.DEFAULT_DATABASE_FILE)
    args = parser.parse_args()

    db = DatabaseV2(args.file)
    db.print_items()



#!/usr/bin/python
# This is from this repo: https://github.com/sharst/seratopy
import logging
import os
import struct
from typing import Tuple, overload

StructType = list[Tuple[str, "ValueType"]]
ValueType = StructType | str | bytes


class Crate(object):
    TRACK_FIELD = "otrk"
    DEFAULT_DATA = [
        ("vrsn", "1.0/Serato ScratchLive Crate"),
        ("osrt", [("tvcn", "key"), ("brev", "\x00")]),
        ("ovct", [("tvcn", "song"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "playCount"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "artist"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "bpm"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "key"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "album"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "length"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "comment"), ("tvcw", "0")]),
        ("ovct", [("tvcn", "added"), ("tvcw", "0")]),
    ]

    def __init__(self, fname):
        self.data: StructType = []

        self.path: str = os.path.dirname(os.path.abspath(fname))
        self.filename: str = os.path.basename(fname)

        # Omit the _Serato_ and Subcrates folders at the end
        self.track_path: str = os.path.join(*Crate._split_path(self.path)[:-2])

        if os.path.exists(fname):
            self.load_from_file(fname)
        else:
            logging.error(f"file does not exist: {fname}. Using default data.")
            self.data = Crate.DEFAULT_DATA

    def __str__(self):
        tracks = self.tracks()
        return f"Crate containing {len(tracks)} tracks: \n{'\n'.join(tracks)}"

    def print_data(self):
        from pprint import pprint

        pprint(self.data)

    @staticmethod
    def _split_path(path: str):
        allparts = []
        while True:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                allparts.insert(0, parts[0])
                break
            elif parts[1] == path:  # sentinel for relative paths
                allparts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                allparts.insert(0, parts[1])
        return allparts

    @staticmethod
    def _get_track_name(value: ValueType) -> str:
        if not isinstance(value, list):
            raise TypeError(f"{Crate.TRACK_FIELD} should be list")
        track_name = value[0][1]
        if not isinstance(track_name, str):
            raise TypeError(f"value should be str")
        return track_name

    def tracks(self) -> list[str]:
        track_names: list[str] = []
        for dat in self.data:
            if dat[0] == Crate.TRACK_FIELD:
                track_name = Crate._get_track_name(dat[1])
                track_names.append(track_name)
        return track_names

    def remove_track(self, track_name: str):
        # track_name needs to include the containing folder name.
        found = False
        for i, dat in enumerate(self.data):
            if dat[0] == Crate.TRACK_FIELD:
                crate_track_name = Crate._get_track_name(dat[1])
                if crate_track_name == track_name:
                    self.data.pop(i)
                    found = True

        if not found:
            raise ValueError(f"Track not found in crate: {track_name}")

    def add_track(self, track_name: str):
        # track name must include the containing folder name
        track_name = os.path.relpath(
            os.path.join(self.track_path, track_name), self.track_path
        )

        if track_name in self.tracks():
            return

        self.data.append((Crate.TRACK_FIELD, [("ptrk", track_name)]))

    def include_tracks_from_folder(self, folder_path: str, replace: bool = False):
        folder_tracks = os.listdir(folder_path)
        folder_tracks = [
            os.path.join(os.path.split(folder_path)[1], track)
            for track in folder_tracks
        ]

        if replace:
            for track in self.tracks():
                if track not in folder_tracks:
                    self.remove_track(track)

        for mfile in folder_tracks:
            self.add_track(mfile)

    class DataTypeError(Exception):
        def __init__(self, data: ValueType, expected_type: type, tag: str | None):
            super().__init__(
                f"data must be {expected_type.__name__} when tag is {tag} (type: {type(data).__name__})"
            )

    @staticmethod
    def _encode(data: ValueType, tag: str | None = None) -> bytes:
        if tag == None or tag[0] == "o":  # struct
            if not isinstance(data, list):
                raise Crate.DataTypeError(data, list, tag)
            ret_data = bytes()
            for dat in data:
                tag = dat[0]
                value = Crate._encode(dat[1], tag=tag)
                length = struct.pack(">I", len(value))
                ret_data = ret_data + tag.encode("utf-8") + length + value
            return ret_data
        elif tag == "vrsn" or tag[0] in ["t", "p"]:  # version or text
            if not isinstance(data, str):
                raise Crate.DataTypeError(data, str, tag)
            return data.encode("utf-16-be")
        elif tag == "sbav" or tag[0] == "b":  # signed or bytes
            if isinstance(data, str):
                return data.encode("utf-8")
            else:
                if not isinstance(data, bytes):
                    raise Crate.DataTypeError(data, bytes, tag)
                return data
        elif tag[0] == "u":  #  unsigned
            return struct.pack(">I", data)
        else:
            raise ValueError(f"unexpected value for tag: {tag}")

    @overload
    @staticmethod
    def _decode(data: bytes, tag: None = None) -> StructType: ...
    @overload
    @staticmethod
    def _decode(data: bytes, tag: str) -> ValueType: ...
    @staticmethod
    def _decode(data: bytes, tag: str | None = None) -> ValueType | StructType:
        if tag == None or tag[0] == "o":  #  struct
            ret_data: StructType = []
            i = 0
            while i < len(data):
                tag = data[i : i + 4].decode("ascii")
                length = struct.unpack(">I", data[i + 4 : i + 8])[0]
                value = data[i + 8 : i + 8 + length]
                value = Crate._decode(value, tag=tag)
                ret_data.append((tag, value))
                i += 8 + length
            return ret_data
        elif tag == "vrsn" or tag[0] in ["t", "p"]:  # version or text
            return data.decode("utf-16-be")
        elif tag == "sbav" or tag[0] == "b":  # signed or bytes
            return data
        elif tag[0] == "u":  # unsigned
            ret_val: bytes = struct.unpack(">I", data)[0]
            return ret_val
        else:
            raise ValueError(f"unexpected value for tag: {tag}")

    def load_from_file(self, fname: str):
        with open(fname, "rb") as mfile:
            self.data = Crate._decode(mfile.read())

    def save_to_file(self, fname: str | None = None):
        if fname is None:
            fname = os.path.join(self.path, self.filename)

        enc_data = Crate._encode(self.data)
        with open(fname, "wb") as mfile:
            mfile.write(enc_data)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("-f", "--filenames_only", action="store_true")
    parser.add_argument(
        "-o", "--output", "--output_file", dest="output_file", default=None
    )
    args = parser.parse_args()

    crate = Crate(args.file)
    tracks = crate.tracks()
    if args.filenames_only:
        track_names = [
            os.path.splitext(os.path.basename(track))[0] for track in crate.tracks()
        ]
        print("\n".join(track_names))
    else:
        print(crate)

    if args.output_file:
        crate.save_to_file(args.output_file)
